# style <name> [= <name>]
# {
#   <option>
# }
#
# widget <widget_set> style <style_name>
# widget_class <widget_class_set> style <style_name>
# accelerator <widget_name> <accelerator>

style "progressbar"
{
  bg[PRELIGHT] = { 22500, 53280, 22500 }	# green
}

style "curve"
{
  fg[NORMAL] = { 58000, 0, 0 }			# red
}

widget "*GtkCurve" style "curve"
widget "*GtkProgressBar" style "progressbar"
